import collections
import logging
import re

import six
import pytimeparse
import schedule

from .exceptions import ConfigurationError

logger = logging.getLogger(__name__)

TimelineRule = collections.namedtuple("TimelineRule", "interval unit at")


class Timeline(object):
    RE_TIME = re.compile(r'\d?\d:\d\d')

    def __init__(self, scheduler=None):
        self.scheduler = scheduler or schedule.default_scheduler

    @classmethod
    def parse_check(cls, check):
        check_schedule = []
        if "period" not in check and "schedule" not in check:
            check_schedule.append(TimelineRule(
                interval=300,
                unit='seconds',
                at=None
            ))
        elif "period" in check:
            period = check["period"]
            if isinstance(period, six.string_types):
                seconds = int(pytimeparse.parse(period))
                logger.debug('Parsed "%s" to %d seconds', period, seconds)
                period = seconds
            check_schedule.append(TimelineRule(
                interval=period,
                unit='seconds',
                at=None
            ))
        if "schedule" in check:
            schedule_value = check['schedule']
            if isinstance(schedule_value, dict):
                check_schedule.append(
                    cls._clean_single_schedule(check, schedule_value)
                )
            elif isinstance(schedule_value, list):
                check_schedule.extend([
                    cls._clean_single_schedule(check, item)
                    for item in schedule_value
                ])
            else:
                raise ConfigurationError(
                    'Check {0} has invalid schedule configuration: {1}'
                    .format(check['name'], check['schedule'])
                )
        return check_schedule

    @classmethod
    def _clean_single_schedule(cls, check, schedule_dict):
        if not isinstance(schedule_dict, dict):
            raise ConfigurationError(
                'Check {0} has invalid schedule configuration: {1}'
                .format(check['name'], schedule_dict),
            )
        try:
            every = schedule_dict['every']
        except KeyError:
            raise ConfigurationError(
                "Check {0} has invalid schedule format: {1}"
                .format(check['name'], schedule_dict)
            )
        if isinstance(every, six.string_types):
            # "every: day" shortcut
            unit, every = every, 1
        else:
            unit = schedule_dict.get('unit')
        unit = cls._clean_unit(check, unit)
        at = cls._clean_at(check, schedule_dict.get('at'))
        rule = TimelineRule(every, unit, at)
        logger.debug('Parsed schedule "%r" to %r', schedule_dict, rule)
        return rule

    @staticmethod
    def _clean_unit(check, unit):
        try:
            getattr(schedule.every(1), unit)
        except:
            raise ConfigurationError(
                "Unit {0} is not one of valid options. Referenced in check {1}"
                .format(unit, check['name'])
            )
        return unit

    @classmethod
    def _clean_at(cls, check, at):
        if at is None or cls.RE_TIME.match(at):
            return at
        raise ConfigurationError(
            'Check {0} schedule has invalid value for "at": {1}'
            .format(check['name'], at)
        )

    def schedule_checks(self, checkers):
        self.scheduler.clear()
        for checker in checkers:
            conf = checker.conf
            for s in conf['schedule']:
                job = getattr(self.scheduler.every(s.interval), s.unit)
                if s.at is not None:
                    logger.info(
                        "Scheduling checks for %r every %r %s at %r",
                        conf["name"],
                        s.interval,
                        s.unit,
                        s.at
                    )
                    job.at(s.at)
                else:
                    logger.info(
                        "Scheduling checks for %r every %r %s",
                        conf["name"],
                        s.interval,
                        s.unit,
                    )
                # Add job to scheduler
                job.do(checker.check)


# The following methods are shortcuts for not having to
# create a Timeline instance:

# default_timeline explicitly uses the default scheduler
default_timeline = Timeline()


def parse_check(check):
    return default_timeline.parse_check(check)


def run_pending():
    return default_timeline.scheduler.run_pending()


def schedule_checks(checkers):
    default_timeline.schedule_checks(checkers)



# -*- coding: utf-8 -*-
from bs4 import BeautifulSoup

from kibitzr.transformer import transform_factory


def run_transform(key, value, content):
    pipeline = transform_factory({'transform': [{key: value}]})
    return pipeline(True, content)


def test_tag_selector():
    ok, content = run_transform('tag', 'a', HTML)
    assert ok is True
    assert content == '<a href="page.html" id="page-link">Page</a>'


def test_css_selector():
    ok, content = run_transform('css', 'body h2.nav a#page-link', HTML)
    assert ok is True
    assert content == '<a href="page.html" id="page-link">Page</a>'


def test_css_selector_all():
    ok, content = run_transform('css-all', 'div', HTML)
    assert ok is True
    assert prettify(content) == u'\n'.join([
        u'<div id="content">',
        u' Привет, Мир!',
        u'</div>',
        u'',
        u'<div class="footer">',
        u' Footer content',
        u'</div>',
    ])


def test_xpath_selector():
    ok, content = run_transform('xpath', '//body/h2/a[@id="page-link"]', HTML)
    assert ok is True
    assert content == '<a href="page.html" id="page-link">Page</a>'


def test_xpath_selector_all():
    ok, content = run_transform('xpath-all', '//div', HTML)
    assert ok is True
    assert content == u'\n'.join([
        u'<div id="content"> Привет, Мир! </div>',
        u'<div class="footer"> Footer content </div>',
    ])


def test_xpath_selector_boolean():
    ok, content = run_transform('xpath', 'boolean(//body/h2/a)', HTML)
    assert ok is True
    assert content == 'True'


def test_xpath_selector_float():
    ok, content = run_transform('xpath', 'count(//div)', HTML)
    assert ok is True
    assert content == '2.0'


def test_xpath_selector_attribute():
    ok, content = run_transform('xpath', '//body/h2/a/@href', HTML)
    assert ok is True
    assert content == 'page.html'


def test_xpath_selector_namespace():
    ok, content = run_transform('xpath', '/html/namespace::*[name()]', HTML)
    assert ok is True
    assert content == 'xml="http://www.w3.org/XML/1998/namespace"'


def test_extract_test():
    ok, content = run_transform('text', None, HTML)
    assert ok is True
    assert content.strip() == u"\n".join([
        u"Page",
        u"Привет, Мир!",
        u"Footer content",
    ])


HTML = u"""<?xml version="1.0" encoding="utf-8"?>
<html>
    <body>
        <h2 class="header nav">
            <a href="page.html" id="page-link">Page</a>
        </h2>
        <div id="content">
            Привет, Мир!
        </div>
        <div class="footer">
            Footer content
        </div>
    </body>
</html>
"""


def prettify(html):
    return "\n".join(
        child.prettify()
        for child in BeautifulSoup(html, "lxml").html.body.children
    )



import pytest

from kibitzr.fetcher.script import fetch_by_script

from ...helpers import stash_mock


def run_script(script):
    return fetch_by_script({'script': script})


def test_fetch_by_script_default_is_bash():
    ok, content = run_script('echo hello')
    assert ok is True
    assert content.strip() == "hello"


def test_fetch_by_explicit_bash_script():
    ok, content = run_script({'bash': 'echo "hello"'})
    assert ok is True
    assert content == "hello\n"


def test_fetch_by_python_script():
    ok, content = run_script({'python': 'ok, content = True, "hello"'})
    assert ok is True
    assert content == "hello"


def test_python_script_must_define_content():
    with pytest.raises(AssertionError):
        run_script({'python': 'dummy'})


def test_python_script_exception_is_reported():
    ok, content = run_script({'python': 'content = 1 / 0'})
    assert ok is False
    assert content.splitlines()[-1].startswith("ZeroDivisionError")


def test_ok_is_optional_in_python_script():
    ok, content = run_script({'python': 'content = "dummy"'})
    assert ok is True
    assert content == "dummy"


def test_bash_error_is_returned():
    ok, content = run_script('no command not found')
    assert ok is False
    assert content.split(':', 1)[1].strip() == "line 1: no: command not found"


def test_python_script_has_stash():
    with stash_mock() as stash:
        stash.write({'key': 'value'})
        ok, content = run_script({'python': 'content = stash["key"]'})
        assert ok is True
        assert content == "value"



"""
FireFox housekeeping - starting and stopping process
"""

import os
import shutil
import logging
from contextlib import contextmanager


PROFILE_DIR = 'firefox_profile'
logger = logging.getLogger(__name__)
FIREFOX_INSTANCE = {
    'headless': None,
    'headed': None,
}


def cleanup():
    """Must be called before exit"""
    temp_dirs = []
    for key in ('headless', 'headed'):
        if FIREFOX_INSTANCE[key] is not None:
            if FIREFOX_INSTANCE[key].profile:
                temp_dirs.append(FIREFOX_INSTANCE[key].profile.profile_dir)
            try:
                FIREFOX_INSTANCE[key].quit()
                FIREFOX_INSTANCE[key] = None
            except:
                logger.exception(
                    "Exception occurred in browser cleanup"
                )
    for temp_dir in temp_dirs:
        shutil.rmtree(temp_dir, ignore_errors=True)


@contextmanager
def firefox(headless=True):
    """
    Context manager returning Selenium webdriver.
    Instance is reused and must be cleaned up on exit.
    """
    from selenium import webdriver
    from selenium.webdriver.firefox.options import Options
    if headless:
        driver_key = 'headless'
        firefox_options = Options()
        firefox_options.add_argument('-headless')
    else:
        driver_key = 'headed'
        firefox_options = None
    # Load profile, if it exists:
    if os.path.isdir(PROFILE_DIR):
        firefox_profile = webdriver.FirefoxProfile(PROFILE_DIR)
    else:
        firefox_profile = None
    if FIREFOX_INSTANCE[driver_key] is None:
        FIREFOX_INSTANCE[driver_key] = webdriver.Firefox(
            firefox_profile=firefox_profile,
            firefox_options=firefox_options,
        )
    yield FIREFOX_INSTANCE[driver_key]
